from collections import Counter

import networkx as nx
import pandas as pd

from dhlab.api.nb_search_api import get_df, get_konks
from dhlab.legacy.nbtext import frame, frame_sort
from dhlab.text.nbtokenizer import tokenize


def phrase_plots(
    phrase_sets,
    title="aftenposten",
    fra=1960,
    til=2020,
    step=5,
    rot=0,
    colours=["r", "b", "g", "y", "m", "c"],
):
    df_all = []
    for f in phrase_sets:
        df_all.append(frame(get_df(f, title=title), ", ".join(f)))
    df = pd.concat(df_all, sort=False)
    df.index = df.index.astype(int)
    df = df.sort_index()
    df["bins"] = pd.cut(df.index, range(fra, til, step), precision=0)
    df.groupby("bins").sum().plot(kind="bar", color=colours, figsize=(15, 5), rot=rot)
    return


def phrase_plots_anno(
    phrase_sets, title="aftenposten", fra=1960, til=2020, rot=0, colours=["r", "b", "g"]
):
    df_all = []
    for f in phrase_sets:
        df_all.append(frame(get_df(f, title=title), ", ".join(f)))
    df = pd.concat(df_all, sort=False)
    df.index = df.index.astype(int)
    df = df.sort_index()
    # df['bins'] = pd.cut(df.index, range(fra, til, step), precision=0)
    df.plot(kind="bar", figsize=(15, 5), rot=rot, color=colours)
    return


def graph_from_df(df, threshold=100):
    edges = []
    normalizer = {
        (x, y): df.stack()[(x, x)] * df.stack()[(y, y)] for (x, y) in df.stack().index
    }
    for x, y in df.stack().index:
        if x != y:
            if df.stack()[(x, y)] > threshold:
                edges.append([x, y, df.stack()[(x, y)] / normalizer[(x, y)]])
    G = nx.Graph()
    G.add_weighted_edges_from(edges)
    return G


def get_all_konks(term, urns):
    konks = []
    for u in urns:
        konks += get_konks(u, term)
    return konks


def collocations_from_nb(word, corpus):
    """Get a concordance, and count the words in it.
    Assume konks reside a dataframe with columns 'after' and 'before'"""
    concordance = frame(get_all_konks(word, corpus))
    return frame_sort(
        frame(
            Counter(
                tokenize(
                    " ".join(concordance["after"].values + concordance["before"].values)
                )
            ),
            word,
        )
    )


def count_from_conc(concordance):
    """From a concordance, count the words in it.
    Assume konks reside a dataframe with columns 'after' and 'before'"""
    word = concordance["word"][0]
    return frame_sort(
        frame(
            Counter(
                tokenize(
                    " ".join(concordance["after"].values + concordance["before"].values)
                )
            ),
            word,
        )
    )



from typing import Union

from IPython.display import HTML
from PIL import Image

from dhlab.api.nb_search_api import (
    iiif_manifest,
    load_picture,
    super_search,
    total_search,
)


def pages(urn, scale=800):
    a = iiif_manifest(urn)
    return [
        page["images"][0]["resource"]["@id"].replace("full/full/", "full/0,{s}/")
        for page in a["sequences"][0]["canvases"]
    ]


def get_url(urn, page=1, part=200):
    # urn as digit
    urn = "URN:NBN:no-nb_digibok_" + urn + "_{0:04d}".format(page)
    print(urn)
    url = f"https://www.nb.no/services/image/resolver/{urn}/full/0,{part}/native.jpg"
    return url


def page_urn(urn, page=1):
    # urn as digit
    return "URN:NBN:no-nb_digibok_" + urn + "_{0:04d}".format(page)


def find_urls(term, number=50, page=0, mediatype="bilder"):
    """generates urls from super_search for pictures"""
    x = super_search(term, number, page, mediatype=mediatype)
    try:
        urls = [
            f["_links"]["thumbnail_custom"]["href"]
            for f in x["_embedded"]["items"]
            if f["accessInfo"]["accessAllowedFrom"] == "EVERYWHERE"
            and "thumbnail_custom" in f["_links"]
        ]
    except:
        urls = []
    return urls


def find_urls2(term, number=50, page=0):
    """generates urls from super_search for pictures"""
    x = super_search(term, number, page)
    try:
        urls = [
            f["_links"]["thumbnail_custom"]["href"]
            for f in x["_embedded"]["mediaTypeResults"][0]["result"]["_embedded"][
                "items"
            ]
            if f["accessInfo"]["accessAllowedFrom"] == "EVERYWHERE"
            and "thumbnail_custom" in f["_links"]
        ]
    except:
        urls = [" ... hmm ..."]
    return urls


def get_picture_from_urn(urn: Union[int, str], width: int = 0, height: int = 300):
    """Fetch the Image object with its URN identifier.

    :meta private:

    :param urn: The uniform resource name number
    :param width: Resolution width of the image.
    :param height: Resolution height of the image.
    :return: An :py:class:`~PIL.Image.Image` object.
    """
    meta = iiif_manifest(urn)
    if "error" not in meta:
        if width == 0 and height == 0:
            url = f"https://www.nb.no/services/image/resolver/{urn}/full/full/0/native.jpg"
        else:
            url = f"https://www.nb.no/services/image/resolver/{urn}/full/{width},{height}/0/native.jpg"
    return Image.open(load_picture(url))


def get_picture_from_url(url, width=0, height=300):
    return Image.open(load_picture(url.format(width=width, height=height)))


def get_metadata_from_url(url):
    import re

    urn = re.findall("(URN.*?)(?:/)", url)[0]
    triple = iiif_manifest(urn)
    r = dict()
    if "error" not in triple:
        r = {x["label"]: x["value"] for x in triple["metadata"] if "label" in x}
    else:
        r = triple["error"]
    return r


def find_urns(term):
    """From result of super_search, to be fed into iiif_manifest"""

    ss = super_search(term)
    urns = [
        f["metadata"]["identifiers"]["urn"]
        for f in ss["_embedded"]["mediaTypeResults"][0]["result"]["_embedded"]["items"]
        if "urn" in f["metadata"]["identifiers"]
    ]
    return urns


def total_urls(number=50, page=0):
    """find urls sequentially"""
    x = total_search(number, page)
    try:
        urls = [
            f["_links"]["thumbnail_custom"]["href"]
            for f in x["_embedded"]["items"]
            if f["accessInfo"]["accessAllowedFrom"] == "EVERYWHERE"
            and "thumbnail_custom" in f["_links"]
        ]
    except:
        urls = []
    return urls


def json2html(meta):
    items = [
        "<dt>{key}</dt><dd>{val}</dd>".format(key=key, val=meta[key]) for key in meta
    ]
    result = "<dl>{items}</dl>".format(items=" ".join(items))
    return result


def display_finds(r):
    """A list of urls in r is displayed as HTML"""
    rows = [
        "<tr><td><img src='{row}'</td><td>{meta}</td></tr>".format(
            row=row, meta=json2html(get_metadata_from_url(row))
        ).format(width=0, height=200)
        for row in r
    ]
    return HTML(
        """<html><head></head>
     <body>
     <table>
     {rows}
     </table>
     </body>
     </html>
     """.format(
            rows=" ".join(rows)
        )
    )



"""Fetch raw files from Github and write them to working directory."""

import contextlib
import os
from pathlib import Path

import requests
from IPython.display import Markdown, display


def printmd(S):
    """Use ``IPython.display`` to render markdown text."""
    display(Markdown(S))
    return


def download_from_github(
    filename=None,
    user=None,
    repository=None,
    branch="master",
    overwrite=False,
    silent=False,
):
    """Fetch a file from Github and write it to working directory.

    Args:
        filename: Filename, including file extension (e.g. `.py` or `.txt`)
        user: Github username of the repo owner.
        repository: Github repository name.
        branch: Name of the repo branch. Defaults to 'master'.
        overwrite: Whether to overwrite existing files in working
            directory. Defaults to not overwrite.
        silent: Whether to output logging messages to stdout.
    """

    nba = requests.get(
        f"https://raw.githubusercontent.com/{user}/{repository}/{branch}/{filename}",
        headers={"Cache-Control": "no-cache"},
    )
    if nba.status_code == 200:
        file_exists = os.path.exists(filename)
        if file_exists and not overwrite:
            if not silent:
                printmd(
                    f"File {os.path.abspath(filename)} exists - call `download_from_github('{filename}', overwrite "
                    f"= True)` in order to download module `{filename}` anyway"
                )
        else:
            with open(filename, "w+", encoding="utf-8") as pyfile:
                pyfile.write(nba.text)
                pyfile.flush()
                pyfile.close()
            if not silent:
                printmd(f"Downloaded file `{os.path.abspath(filename)}`")
    else:
        printmd(f"Failed to download {filename} with http code {nba.status_code}")


def get_file_from_github(url, overwrite=False, silent=False):
    """Fetch a file on github.

    it is enough with reference
    it will look in raw user content for the file.

    Args:
        overwrite: defaults to no overwrite
        silent: default is not silent
    """

    if url.startswith("https://github.com/") or url.startswith("github.com"):
        trail = url.split("github.com")[-1].replace("blob/", "")
        fileref = f"""https://raw.githubusercontent.com{trail}"""
    elif url.startswith("raw.githubusercontent"):
        fileref = f"""https://{url}"""
    elif url.startswith("https://raw.githubusercontent"):
        fileref = url
    else:
        return "is this is a file on github?"

    nba = requests.get(f"{fileref}", headers={"Cache-Control": "no-cache"})
    if nba.status_code == 200:
        filename = os.path.basename(fileref)
        file_exists = os.path.exists(filename)
        if file_exists and not overwrite:
            if not silent:
                printmd(
                    f"File {os.path.abspath(filename)} exists -"
                    f" call `download('{filename}', overwrite = True)` in order to "
                    f"download `{filename}` anyway"
                )
        else:
            nba = nba.text
            with open(filename, "w", encoding="UTF-8") as pyfile:
                pyfile.write(nba)
                pyfile.flush()
                pyfile.close()
            if not silent:
                printmd(f"Downloaded file `{os.path.abspath(filename)}`")
    else:
        printmd(
            f"Failed to download {fileref} with http response code {nba.status_code}"
        )

    return


@contextlib.contextmanager
def working_directory(path):
    """Changes working directory and returns to previous on exit.

    Source: https://stackoverflow.com/a/42441759

    :meta private:
    """
    prev_cwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(prev_cwd)



import os
import re

import requests
from IPython.display import display, HTML, Markdown
from PIL import Image

small_scale = 0.59
large_scale = 1.58


def nb_search(
    term="",
    creator="",
    number=50,
    page=0,
    mediatype="bøker",
    lang="nob",
    period=(18000101, 20401231),
    extra_filters=None,
    extra_conditions=None,
):
    """Søk etter term og få ut json"""
    number = min(number, 50)

    filters = []
    aq = []

    if extra_filters is not None:
        filters += extra_filters
    params = {"page": page, "size": number}
    if extra_conditions is not None:
        for x in extra_conditions:
            params[x] = extra_conditions[x]

    if lang != "":
        aq.append(f"languages:{lang}")

    if creator != "":
        filters.append(f"creator:{creator}")

    if mediatype != "":
        filters.append(f"mediatype:{mediatype}")

    if period != ():
        filters.append(f"date:[{period[0]} TO {period[1]}]")

    if filters:
        params["filter"] = filters

    if aq:
        params["aq"] = aq

    if term != "":
        params["q"] = term

    r = requests.get("https://api.nb.no:443/catalog/v1/items", params=params)
    return r.json()


def div_wrapper(div_content):
    return f"""<div>{div_content}</div>"""


def display_books(books, width=100):
    """A dictionary of urns - urls is displayed"""

    def html_wrapper(body):
        return f"""<style>
    img {{
        width:{width}px;
        height:auto;
        max-width:100%}}
    </style>
    <body>{body}</body>"""

    book_divs = ""
    for u in books:
        mf = iiif_manifest(u)
        thumbnail = f"<img src = '{mf['thumbnail']['@id']}'></img>"
        metainfo = "\n".join(
            [f"<b>{x['label']}</b>{x['value']}" for x in mf["metadata"]]
        )
        imgs = "\n".join([f"<img src='{pic_url}'></img>" for pic_url in books[u]])
        book_divs += div_wrapper(thumbnail + metainfo + imgs)
    return html_wrapper(book_divs)


def markdown_books(books, width=100):
    """A dictionary of urns - urls is displayed"""

    book_divs = ""
    for u in books:
        mf = iiif_manifest(u)
        thumbnail = f"<h3>Forside</h3> <img src='{mf['thumbnail']['@id']}'></img>\n\n"
        metainfo = "\n".join(
            [
                f"<h3>Metadata</h3> <b>{x['label']}</b> {x['value']}"
                for x in mf["metadata"]
            ]
        )
        imgs = "\n".join(
            [
                f"<img style='float:right' src='{pic_url}' width = {width}></img>"
                for pic_url in books[u]
            ]
        )
        book_divs += div_wrapper(thumbnail + metainfo + imgs)
    return book_divs


def iiif_manifest(urn):
    if "URN" not in str(urn) and "digibok" not in str(urn):
        urn = "URN:NBN:no-nb_digibok_" + str(urn)
    elif "URN" not in str(urn):
        urn = "URN:NBN:no-nb_" + str(urn)
    r = requests.get(f"https://api.nb.no/catalog/v1/iiif/{urn}/manifest")
    return r.json()


def urns_from_super(
    term="",
    creator="",
    number=50,
    page=0,
    mediatype="bøker",
    lang="nob",
    period=(18000101, 20401231),
    filters=None,
    conditions=None,
):
    res = nb_search(
        term,
        mediatype=mediatype,
        creator=creator,
        number=number,
        page=page,
        lang=lang,
        period=period,
        extra_filters=filters,
        extra_conditions=conditions,
    )
    ids = [x["metadata"]["identifiers"] for x in res["_embedded"]["items"]]
    return [x["urn"] for x in ids if "urn" in x]


def get_illustration_data_from_book(urn):
    if "digibok" in str(urn):
        urn = re.findall("[0-9]{13}", str(urn))[0]
    r = requests.get("https://api.nb.no/ngram/illustrations", json={"urn": urn})
    return r.json()


def get_urls_from_illustration_data(illus, part=True, scale=None, cuts=True, delta=0):
    """
    Restricted images must not go over 1024 x 1024 pixels.

    :param illus: dictionary of this format:

        ``{'height': 270, 'hpos': 251, 'page': 'digibok_2017081626006_0018',
        'resolution': 400, 'vpos': 791, 'width': 373}``

    :param part: If True, return a cut out of the image
    :param scale: To be filled in
    :param cuts: If true, allow cropping of image
    :param delta: To be filled in
    :return: URL of the image
    """
    if scale is None:
        if illus["resolution"] >= 300 or illus["resolution"] < 100:
            scale = large_scale
        else:
            scale = small_scale

    height = int(illus["height"]) + 2 * delta
    width = int(illus["width"]) + 2 * delta
    vpos = int(illus["vpos"]) - delta
    hpos = int(illus["hpos"]) - delta

    if cuts:
        if width * scale > 1024:
            width = int(1024 / scale)
        if height * scale > 1024:
            height = int(1024 / scale)

    urn = "URN:NBN:no-nb_" + illus["page"]
    url_prefix = "https://www.nb.no/services/image/resolver"
    if part:
        # return cut out
        url = (
            f"{url_prefix}/"
            f"{urn}/"
            f"{int(hpos * scale)},"
            f"{int(vpos * scale)},"
            f"{int(width * scale)},"
            f"{int(height * scale)}"
            f"/full/0/native.jpg"
        )
    else:
        # return whole page
        url = f"{url_prefix}/" f"{urn}/" f"full/" f"0,{part}/" f"0/native.jpg"

    return url


def show_illustrations_urn(urn, tilgjengelig="fritt"):
    display(
        Markdown(
            "\n".join(
                [
                    "**" + x["label"] + "**: " + x["value"]
                    for x in iiif_manifest(urn)["metadata"]
                ]
            )
        )
    )

    if tilgjengelig.lower().startswith("fri"):
        c = False
    else:
        c = True
    return display_finds(
        [
            get_urls_from_illustration_data(u, cuts=c)
            for u in get_illustration_data_from_book(urn)
        ]
    )


def display_finds_meta(r):
    """A list of urls in r is displayed as HTML"""
    rows = [
        (
            f"<tr><td><img src='{row}'</td><td><a href = {row} target='_'>"
            f"{row}</a></td></tr>"
        )
        for row in r
    ]
    return HTML(
        f"""<html><head></head>
     <body>
     <table>
     {' '.join(rows)}
     </table>
     </body>
     </html>
     """
    )


def display_finds(r):
    """A list of urls in r is displayed as HTML"""
    rows = [f"<tr><td><img src='{row}'</td></tr>" for row in r]
    return HTML(
        f"""<html><head></head>
     <body>
     <table>
     {' '.join(rows)}
     </table>
     </body>
     </html>
     """
    )


def url2urn(url):
    return re.findall("URN:.*[0-9]{13}", url)[0]


def mods(urn):
    r = requests.get(f"https://api.nb.no:443/catalog/v1/metadata/{urn}/mods")
    return r.json()


def pages(urn, scale=800):
    a = iiif_manifest(urn)
    return [
        page["images"][0]["resource"]["@id"].replace("full/full/", f"full/0,{scale}/")
        for page in a["sequences"][0]["canvases"]
    ]


def load_picture(url):
    r = requests.get(url, stream=True)
    r.raw.decode_content = True
    return r.raw


def get_url(urn, page=1, part=200):
    """From a representaion of URN mapped to digibok.

    A URN is a serial number or any string with serial number.
    """
    urnserial = re.findall("[0-9]+", str(urn))
    if urnserial:
        urnserial = urnserial[0]
    else:
        return ""
    urn = f"URN:NBN:no-nb_digibok_{urnserial}_{page:04d}"
    print(urn)
    url_prefix = "https://www.nb.no/services/image/resolver"
    url = f"{url_prefix}/{urn}/full/0,{part}/0/native.jpg"
    return url


def page_urn(urn, page=1):
    # urn as digit
    return f"URN:NBN:no-nb_digibok_{urn}_{page:04d}"


def super_search(term, number=50, page=0, mediatype="bilder"):
    """Søk etter term og få ut json"""
    number = min(number, 50)
    if term == "":
        r = requests.get(
            "https://api.nb.no:443/catalog/v1/items",
            params={"filter": f"mediatype:{mediatype}", "page": page, "size": number},
        )
    else:
        r = requests.get(
            "https://api.nb.no:443/catalog/v1/items",
            params={
                "q": term,
                "filter": f"mediatype:{mediatype}",
                "page": page,
                "size": number,
            },
        )
    return r.json()


def find_urls(term, number=50, page=0, mediatype="bilder"):
    """generates urls from super_search for pictures"""
    x = super_search(term, number, page, mediatype=mediatype)
    try:
        urls = [
            f["_links"]["thumbnail_custom"]["href"]
            for f in x["_embedded"]["items"]
            if f["accessInfo"]["accessAllowedFrom"] == "EVERYWHERE"
            and "thumbnail_custom" in f["_links"]
        ]
    except (
        BaseException
    ):  # PEP8: E722 do not use bare 'except', too broad exception clause
        urls = []
    return urls


def find_urls2(term, number=50, page=0):
    """generates urls from super_search for pictures"""
    x = super_search(term, number, page)
    try:
        urls = [
            f["_links"]["thumbnail_custom"]["href"]
            for f in x["_embedded"]["mediaTypeResults"][0]["result"]["_embedded"][
                "items"
            ]
            if f["accessInfo"]["accessAllowedFrom"] == "EVERYWHERE"
            and "thumbnail_custom" in f["_links"]
        ]
    except (
        BaseException
    ):  # PEP8: E722 do not use bare 'except', too broad exception clause
        urls = [" ... hmm ..."]
    return urls


def get_picture_from_urn(urn, width=0, height=300):
    meta = iiif_manifest(urn)
    url_prefix = "https://www.nb.no/services/image/resolver"
    if "error" not in meta:
        if width == 0 and height == 0:
            url = f"{url_prefix}/{urn}/full/full/0/native.jpg"
        else:
            url = f"{url_prefix}/{urn}/full/{width},{height}/0/native.jpg"
    return Image.open(load_picture(url))


def get_picture_from_url(url, width=0, height=300):
    return Image.open(load_picture(url.format(width=width, height=height)))


def get_metadata_from_url(url):
    urn = re.findall("(URN.*?)(?:/)", url)[0]
    triple = iiif_manifest(urn)
    r = {}  # Local variable 'r' value is not used
    if "error" not in triple:
        r = {x["label"]: x["value"] for x in triple["metadata"] if "label" in x}
    else:
        r = triple["error"]
    return r


def find_urns(term):
    """From result of super_search, to be fed into iiif_manifest."""

    ss = super_search(term)
    urns = [
        f["metadata"]["identifiers"]["urn"]
        for f in ss["_embedded"]["mediaTypeResults"][0]["result"]["_embedded"]["items"]
        if "urn" in f["metadata"]["identifiers"]
    ]
    return urns


def total_search(size=50, page=0):
    """Finn de første antallet = 'size' fra side 'page' og få ut json"""
    size = min(size, 50)
    r = requests.get(
        "https://api.nb.no:443/catalog/v1/items",
        params={"filter": "mediatype:bilder", "page": page, "size": size},
    )
    return r.json()


def total_urls(number=50, page=0):
    """find urls sequentially"""
    x = total_search(number, page)
    try:
        urls = [
            f["_links"]["thumbnail_custom"]["href"]
            for f in x["_embedded"]["items"]
            if f["accessInfo"]["accessAllowedFrom"] == "EVERYWHERE"
            and "thumbnail_custom" in f["_links"]
        ]
    except (
        BaseException
    ):  # PEP8: E722 do not use bare 'except', too broad exception clause
        urls = []
    return urls


def find_urns_sesam(term="", creator="", number=50, page=0, mediatype="bilder"):
    """generates urls from super_search for pictures"""
    x = nb_search(
        term=term, creator=creator, number=number, page=page, mediatype=mediatype
    )
    try:
        sesamid = [
            f["id"]
            for f in x["_embedded"]["items"]
            if f["accessInfo"]["accessAllowedFrom"] == "EVERYWHERE"
            and "thumbnail_custom" in f["_links"]
        ]
    except BaseException:
        sesamid = []
    return sesamid


def save_pictures(pages_, urn, root="."):
    """Save picture references in pages on the form::

        pages = {
            urn1: [page1, page2, ..., pageN],
            urn2: [page1, ..., pageM],
            ...
            urnK: [page1, ..., pageL]}

    Parameter urn is one of the keys in pages,
    where each page reference is a URL.
    """

    # In case urn is an actual URN, works also if urn is passed as sesamid

    folder_name = urn.split(":")[-1]
    folder_ref = os.path.join(root, folder_name)
    try:
        os.mkdir(folder_ref)

    except FileExistsError:
        pass

    for p in pages_[urn]:
        # pell ut entydig referanse til bildet fra URL-en i bildelisten som
        # filnavn

        filename = p.split("/")[6].split(":")[-1] + ".jpg"

        path = os.path.join(folder_ref, filename)
        get_picture_from_url(p).save(path)

    return True


def save_all_pages(pages_, root="."):
    """Save picture references in pages on the form::

        pages = {
            urn1 : [page1, page2, ..., pageN],
            urn2: [page1, ..., pageM],
            ...
            urnK: [page1, ..., pageL]
        }

    Each page reference is a URL.
    """

    # In case urn is an actual URN, works also if urn is passed as sesamid
    for urn in pages_:
        folder_name = urn.split(":")[-1]
        folder_ref = os.path.join(root, folder_name)
        try:
            os.mkdir(folder_ref)

        except FileExistsError:
            pass

        for p in pages_[urn]:
            # pell ut entydig referanse til bildet fra URL-en i bildelisten som
            # filnavn

            filename = p.split("/")[6].split(":")[-1] + ".jpg"

            path = os.path.join(folder_ref, filename)
            get_picture_from_url(p).save(path)

    return True


def json2html(meta):
    items = [f"<dt>{key}</dt><dd>{value}</dd>" for key, value in meta.items()]
    result = f"<dl>{' '.join(items)}</dl>"
    return result



"""The dhlab python library provides functions and objects
to retrieve and analyze image and text data from the digital collection of the National Library of Norway.

You can build text corpora, retrieve their metadata, search for images, and do quantitative analyses.

The dhlab python package calls the [DHLAB API](https://api.nb.no/dhlab/) under the hood to retrieve data.
"""

# api
from dhlab.api.dhlab_api import totals

# legacy code
from dhlab.legacy import (
    graph_networkx_louvain,
    module_update,
    nb_external_files,
    nbpictures,
    nbtext,
    token_map,
)

# metadata
from dhlab.metadata.natbib import metadata_from_urn, metadata_query

# ngram
from dhlab.ngram.ngram import Ngram, NgramBook, NgramNews

# text
from dhlab.text import nbtokenizer
from dhlab.text.chunking import Chunks
from dhlab.text.conc_coll import Collocations, Concordance, Counts
from dhlab.text.corpus import Corpus
from dhlab.text.dispersion import Dispersion
from dhlab.text.geo_data import GeoData, GeoNames
from dhlab.text.parse import NER, POS, Models
from dhlab.text.wildcard import WildcardWordSearch
from dhlab.utils.display import css
from dhlab.utils.files import download_from_github, get_file_from_github

# wordbank
from dhlab.wordbank.wordbank import WordForm, WordLemma, WordParadigm



import json
from collections import Counter

import matplotlib.pyplot as plt

try:
    from wordcloud import WordCloud
except BaseException:
    print("wordcloud er ikke installert, kan ikke lage ordskyer")


def make_cloud(
    json_text,
    top=100,
    background="white",
    stretch=lambda x: 2 ** (10 * x),
    width=500,
    height=500,
    font_path=None,
):
    """Create a word cloud from a frequency list."""
    pairs0 = Counter(json_text).most_common(top)
    pairs = {x[0]: stretch(x[1]) for x in pairs0}
    wc = WordCloud(
        font_path=font_path,
        background_color=background,
        width=width,
        # color_func=my_colorfunc,
        ranks_only=True,
        height=height,
    ).generate_from_frequencies(pairs)
    return wc


def draw_cloud(sky, width=20, height=20, fil=""):
    """Draw a word cloud produced by :func:`make_cloud`."""
    plt.figure(figsize=(width, height))
    plt.imshow(sky, interpolation="bilinear")
    figplot = plt.gcf()
    if fil != "":
        figplot.savefig(fil, format="png")


def cloud(
    df,
    column="",
    top=200,
    width=1000,
    height=1000,
    background="black",
    file="",
    stretch=10,
    font_path=None,
):
    """Make and draw a cloud from a pandas dataframe, using :func:`make_cloud` and
    :func:`draw_cloud`."""
    if column == "":
        column = df.columns[0]
    data = json.loads(df[column].to_json())
    a_cloud = make_cloud(
        data,
        top=top,
        background=background,
        font_path=font_path,
        stretch=lambda x: 2 ** (stretch * x),
        width=width,
        height=height,
    )
    draw_cloud(a_cloud, fil=file)



from dhlab.api.dhlab_api import ner_from_urn, pos_from_urn, show_spacy_models


class Models:
    """Show the spaCy language models available"""

    def __init__(self):
        self.models = show_spacy_models()


class NER:
    """Provide NER"""

    def __init__(
        self,
        urn: str | None = None,
        model: str | None = None,
        start_page: int = 0,
        to_page: int = 0
    ):
        self.model = model
        self.ner = ner_from_urn(
            urn=urn, model=self.model, start_page=start_page, to_page=to_page
        )


class POS:
    """Provide POS and a parse"""

    def __init__(
        self,
        urn: str | None = None,
        model: str | None = None,
        start_page: int = 0,
        to_page: int = 0
    ):
        self.model = model
        self.pos = pos_from_urn(
            urn=urn, model=self.model, start_page=start_page, to_page=to_page
        )



import pandas as pd
from pandas import DataFrame

from dhlab.api.nb_ngram_api import get_ngram


def nb_ngram(
    terms: str,
    corpus: str = "bok",
    smooth: int = 1,
    years: tuple = (1810, 2010),
    mode: str = "relative",
    lang: str = "nob",
):
    """Extract N-gram frequencies from given `terms` and `years`.
    `lang` param is not supported for corpus=`avis` and will be set to None if `avis` is passed.

    The `lang` param is not supported for `corpus="avis"` and will be set to None if `avis` is passed.

    Returns:
        A sorted Pandas DataFrame indexed by year, with columns for each term.

    :meta private:
    """
    # Set default lang for 'bok'-corpus
    if corpus == "avis":
        lang = None

    df = ngram_conv(
        get_ngram(terms, corpus=corpus, lang=lang),
        smooth=smooth,
        years=years,
        mode=mode,
    )
    df.index = df.index.astype(int)
    return df.sort_index()


## tar tilbake til original den her virker ikke LGJ
def ngram_conv_old(
    ngrams, smooth: int = 1, years: tuple = (1810, 2013), mode: str = "relative"
) -> DataFrame:
    """Construct a dataframe with ngram mean frequencies per year over a given time period.

    :meta private:
    """
    ngc = {}
    # check if relative frequency or absolute frequency is in question
    if mode.startswith("rel") or mode == "y":
        arg = "y"
    else:
        arg = "f"
    for x in ngrams:
        if x and isinstance(x, list):
            ngc[x["key"]] = {
                z["x"]: z[arg]
                for z in x["values"]
                if years[1] >= int(z["x"]) >= years[0]
            }
    return pd.DataFrame(ngc).rolling(window=smooth, win_type="triang").mean()


def ngram_conv(
    ngrams,
    smooth: int = 1,
    years: tuple = (1810, 2013),
    mode: str = "relative",
) -> DataFrame:
    """Construct a dataframe with ngram mean frequencies per year over a given time period.

    Args:
        ngrams: To be filled in.
        smooth: Smoothing factor for the graph visualisation.
        years: Tuple with start and end years for the time period of interest
        mode: Frequency measure.

    Returns:
        pandas dataframe with mean values for each year

    :meta private:
    """
    ngc = {}
    # check if relative frequency or absolute frequency is in question
    if mode.startswith("rel") or mode == "y":
        arg = "y"
    else:
        arg = "f"
    for x in ngrams:
        # check if x is a non empty ngram - empty ngrams are coded as empty lists
        # if x is non emtpy it accepts keys - look at alternative isinstance(x, dict)?
        if x != []:
            ngc[x["key"]] = {
                z["x"]: z[arg]
                for z in x["values"]
                if int(z["x"]) <= int(years[1]) and int(z["x"]) >= int(years[0])
            }

    return pd.DataFrame(ngc).rolling(window=smooth, win_type="triang").mean()



"""Retrieve metadata from objects in the digital archive."""
from typing import List, Optional

import pandas as pd
import requests

from dhlab.constants import BASE_URL


def get_metadata(urns: Optional[List] = None):
    """Fetch metadata from a list of urns.

    Args:
        urns: uniform resource names, example:
            `["URN:NBN:no-nb_digibok_2011051112001"]`
    """
    r = requests.post(f"{BASE_URL}/get_metadata", json={"urns": urns})
    return pd.DataFrame(r.json())



from io import StringIO
from typing import Dict, List, Tuple, Union

import pandas as pd

# from requests import HTTPError, JSONDecodeError, ConnectionError
from pandas import DataFrame, Series, SparseDtype

from dhlab.constants import BASE_URL
from dhlab.api.utils import api_get, api_post, DHLabApiError

pd.options.display.max_rows = 100

# wildcard search for words


def wildcard_search(word: str, factor: int | None = 2, freq_limit: int | None = 10, limit: int | None = 50) -> DataFrame:
    """Get words, with frequencies, using '*' as a wildcard.

    For example, searching "ord*en*" might return:
        ```
                      freq
        ordbogen       874
        ordboken     10604
        ...
        ordningen   368131
        ordnmgen       722
        ...
        ```

    :param word: Word to search, allowing (potentially multiple) '*' as a wildcard
    :param factor: Max length of matched words, as a factor of `word`
    :param freq_limit: Lower frequency limit of returned matched words
    :param limit: Max number of returned results, prioritized by frequency
    """
    resp = api_get(
        f"{BASE_URL}/wildcard_word_search",
        params={"word": word, "factor": factor, "freq_lim": freq_limit, "limit": limit},
    )

    return pd.DataFrame.from_dict(resp.json(), orient="index", columns=["freq"])


# fetch metadata


def images(text: str | None = None, part: int | None = True, hits: int | None = 500, delta: int | None = 0):
    """Retrive images from bokhylla

    :param text: Fulltext query expression for sqlite.
    :param part: If a number, the whole page is shown. If True, get auto-scaled image.
    :param delta: If part==True, show `delta` additional pixels on each side of image
    :param hits: Number of images
    :return: Requests `res.json()` return value
    """

    resp = api_get(
        f"{BASE_URL}/images",
        params = {"text": text, "part": part, "hits": hits, "delta": delta}
    )

    return resp.json()


def ner_from_urn(
    urn: str | None = None, model: str | None = None, start_page: int = 0, to_page: int  = 0
) -> DataFrame:
    """Get NER annotations for a text (``urn``) using a spacy ``model``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :param str model: name of a spacy model.
        Check which models are available with :func:`show_spacy_models`
    :return: Dataframe with annotations and their frequencies
    """
    resp = api_get(
        f"{BASE_URL}/ner_urn",
        params={"urn": urn, "model": model, "start_page": start_page, "to_page": to_page}
    )

    return pd.read_json(resp.json())


def pos_from_urn(
    urn: str | None = None, model: str | None = None, start_page: int = 0, to_page: int = 0
) -> DataFrame:
    """Get part of speech tags and dependency parse annotations for a text (``urn``) with a SpaCy ``model``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :param str model: name of a spacy model.
        Check which models are available with :func:`show_spacy_models`
    :param int start_page:
    :param int to_page:
    :return: Dataframe with annotations and their frequencies
    """
    resp = api_get(
        f"{BASE_URL}/pos_urn",
        params={"urn": urn, "model": model, "start_page": start_page, "to_page": to_page}
    )

    return pd.read_json(resp.json())


def show_spacy_models() -> List:
    """Show available SpaCy model names."""
    resp = api_get(f"{BASE_URL}/ner_models")

    return resp.json()


def get_places(urn: str) -> DataFrame:
    """Look up placenames in a specific URN.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/places <https://api.nb.no/dhlab/#/default/post_places>`_.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    """
    r = api_post(f"{BASE_URL}/places", json={"urn": urn})
    return pd.DataFrame(r.json())


def geo_lookup(
    places: List,
    feature_class: str | None = None,
    feature_code: str | None = None,
    field: str = "alternatename",
) -> DataFrame:
    """From a list of places, return their geolocations

    :param list places: a list of place names - max 1000
    :param str feature_class: which GeoNames feature class to return. Example: ``P``
    :param str feature_code: which GeoNames feature code to return. Example: ``PPL``
    :param str field: which name field to match - default "alternatename".
    """
    res = api_post(
        f"{BASE_URL}/geo_data",
        json={
            "words": places,
            "feature_class": feature_class,
            "feature_code": feature_code,
            "field": field,
        },
    )
    columns = [
        "geonameid",
        "name",
        "alternatename",
        "latitude",
        "longitude",
        "feature_class",
        "feature_code",
    ]
    return pd.DataFrame(res.json(), columns=columns)


def get_dispersion(
    urn: str | None = None,
    words: List | None = None,
    window: int = 300,
    pr: int = 100,
) -> Series:
    """Count occurrences of words in the given URN object.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint ``/dispersion``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :param list words: list of words. Defaults to a list of punctuation marks.
    :param int window: The number of tokens to search through per row. Defaults to 300.
    :param int pr: defaults to 100.
    :return: a ``pandas.Series`` with frequency counts of the words in the URN object.
    """
    params = {"pr": pr, "urn": urn, "window": window, "words": words}
    r = api_post(f"{BASE_URL}/dispersion", json=params)
    return pd.Series(r.json())


def get_metadata(urns: List[str] | None = None) -> DataFrame:
    """Get metadata for a list of URNs.

    Calls the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/get_metadata <https://api.nb.no/dhlab/#/default/post_get_metadata>`_.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    """
    r = api_post(f"{BASE_URL}/get_metadata", json={"urns": urns})
    return DataFrame(r.json())


def get_identifiers(identifiers: list | None = None) -> list:
    """Convert a list of identifiers, oaiid, sesamid, urns or isbn10 to dhlabids"""

    if identifiers is None:
        identifiers = []

    res = api_post(
        f"{BASE_URL}/identifiers",
        json={"identifiers": [i for i in identifiers if i != ""]},
    )
    return res.json()


def get_chunks(urn: str | None = None, chunk_size: int = 300) -> Union[Dict, List]:
    """Get the text in the document ``urn`` as frequencies of chunks
     of the given ``chunk_size``.

    Calls the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    ``/chunks``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :param int chunk_size: Number of tokens to include in each chunk.
    :return: list of dicts with the resulting chunk frequencies, or an empty dict
    """
    resp = api_get(
        f"{BASE_URL}/chunks",
        params={"urn": urn, "chunk_size": chunk_size}
    )

    return resp.json()


def get_chunks_para(urn: str | None = None) -> Union[Dict, List]:
    """Fetch chunks and their frequencies from paragraphs in a document (``urn``).

    Calls the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    ``/chunks_para``.

    :param str urn: uniform resource name, example: ``URN:NBN:no-nb_digibok_2011051112001``
    :return: list of dicts with the resulting chunk frequencies, or an empty dict
    """
    resp = api_get(
        f"{BASE_URL}/chunks_para",
        params={"urn": urn}
    )

    return resp.json()


def evaluate_documents(wordbags: Dict | None = None, urns: List[str] | None = None) -> DataFrame:
    """Count and aggregate occurrences of topic ``wordbags`` for each document in a list of ``urns``.

    :param dict wordbags: a dictionary of topic keywords and lists of associated words.
        Example: ``{"natur": ["planter", "skog", "fjell", "fjord"], ... }``
    :param list urns: uniform resource names, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :return: a ``pandas.DataFrame`` with the topics as columns, indexed by the dhlabids of the
        documents.
    """
    res = api_post(
        f"{BASE_URL}/evaluate", json={"wordbags": wordbags, "urns": urns}
    )
    df = pd.DataFrame(res.json()).transpose()
    return df


def get_reference(
    corpus: str = "digavis",
    from_year: int = 1950,
    to_year: int = 1955,
    lang: str = "nob",
    limit: int = 100000,
) -> DataFrame:
    """Reference frequency list of the n most frequent words from a given corpus in a given period.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/reference_corpus <https://api.nb.no/dhlab/#/default/get_reference_corpus>`_.

    :param str corpus: Document type to include in the corpus, can be either ``'digibok'`` or
        ``'digavis'``.
    :param int from_year: Starting point for time period of the corpus.
    :param int to_year: Last year of the time period of the corpus.
    :param str lang: Language of the corpus, can be one of
        ``'nob,', 'nno,', 'sme,', 'sma,', 'smj', 'fkv'``
    :param int limit: Maximum number of most frequent words.
    :return: A ``pandas.DataFrame`` with the results.
    """
    resp = api_get(
        BASE_URL + "/reference_corpus",
        params={"corpus": corpus, "from_year": from_year, "to_year": to_year, "lang": lang, "limit": limit}
    )

    return pd.DataFrame(resp.json(), columns=["word", "freq"]).set_index("word")


def find_urns(docids: Union[Dict, DataFrame] | None = None, mode: str = "json") -> DataFrame:
    """Return a list of URNs from a collection of docids.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/find_urn`.

    :param docids: dictionary of document IDs (``{docid: URN}``) or a ``pandas.DataFrame``.
    :param str mode: Default 'json'.
    :return: the URNs that were found, in a ``pandas.DataFrame``.
    """
    resp = api_post(
        BASE_URL + "/find_urn",
        json={"docids": docids, "mode": mode}
    )

    return pd.DataFrame.from_dict(resp.json(), orient="index", columns=["urn"])


def _ngram_doc(
    doctype: str = "",
    word: List | str | None = None,
    title: str | None = None,
    period: Tuple[int, int] | None = None,
    publisher: str | None = None,
    lang: str | None = None,
    city: str | None = None,
    ddk: str | None = None,
    topic: str | None = None,
) -> DataFrame:
    """Count occurrences of one or more words over a time period.

    The type of document to search through is decided by the `doctype`.
    Filter the selection of documents with metadata.
    Use % as wildcard where appropriate - no wildcards in `word` or `lang`.

    Args:
        doctype: API endpoint for the document type to get ngrams for.
            Can be `'book'`, `'periodicals'`, or `'newspapers'`.
        word: Word(s) to search for.
            Can be several words in a single string, separated by comma, e.g. `"ord,ordene,orda"`.
        title: Title of a specific document to search through.
        period: Start and end years or dates of a time period,
            given as `(YYYY, YYYY)`` or `(YYYYMMDD, YYYYMMDD)`.
        publisher: Name of a publisher.
        lang: Language as a 3-letter ISO code (e.g. `"nob"` or `"nno"`)
        city: City of publication.
        ddk: [Dewey Decimal Classification](https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon) identifier.
        topic: Topic of the documents.

    Returns:
        a `pandas.DataFrame` with the resulting frequency counts of the word(s),
            spread across years. One year per row.
    """
    if word is None:
        word = ["."]

    params = {"doctype": doctype, "word": word, "title": title, "period": period,
              "publisher": publisher, "lang": lang, "city": city, "ddk": ddk, "topic": topic}
    if isinstance(word, str):
        # assume a comma separated string
        word = [w.strip() for w in word.split(",")]
    params["word"] = tuple(word)
    params = {x: params[x] for x in params if params[x] is not None}
    r = api_post(BASE_URL + "/ngram_" + doctype, json=params)
    df = pd.DataFrame.from_dict(r.json(), orient="index")
    df.index = df.index.map(lambda x: tuple(x.split()))
    if not isinstance(df.index, pd.MultiIndex):
        raise DHLabApiError(f"{isinstance(df.index, pd.MultiIndex)=}")
    columns = df.index.levels[0]
    df = pd.concat([df.loc[x] for x in columns], axis=1)
    df.columns = columns
    return df


def reference_words(
    words: List | None = None,
    doctype: str = "digibok",
    from_year: Union[str, int] = 1800,
    to_year: Union[str, int] = 2000,
) -> DataFrame:
    """Collect reference data for a list of words over a time period.

    Reference data are the absolute and relative frequencies of the ``words``
    across all documents of the given ``doctype`` in the given time period
    (``from_year`` - ``to_year``).

    :param list words: list of word strings.
    :param str doctype: type of reference document. Can be ``"digibok"`` or ``"digavis"``.
        Defaults to ``"digibok"``.

        .. note::
           If any other string is given as the ``doctype``,
           the resulting data is equivalent to what you get with
           ``doctype="digavis"``.

    :param int from_year: first year of publication
    :param int to_year: last year of publication
    :return: a DataFrame with the words' frequency data
    """
    resp = api_post(
        f"{BASE_URL}/reference_words",
        json={"words": words, "doctype": doctype, "from_year": from_year, "to_year": to_year}
    )

    return pd.DataFrame(resp.json(), columns=["word", "freq", "relative"])


# @_docstring_parameters_from(_ngram_doc, drop="doctype")
def ngram_book(
    word: Union[List, str] = ["."],
    title: str | None = None,
    period: Tuple[int, int] | None = None,
    publisher: str | None = None,
    lang: str | None = None,
    city: str | None = None,
    ddk: str | None = None,
    topic: str | None = None,
) -> DataFrame:
    """Count occurrences of one or more words in books over a given time period.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/ngram_book`.

    Filter the selection of books with metadata.
    Use % as wildcard where appropriate - no wildcards in ``word`` or ``lang``.

    :param word: Word(s) to search for.
        Can be several words in a single string, separated by comma, e.g. ``"ord,ordene,orda"``.
    :type word: str or list of str
    :param str title: Title of a specific document to search through.
    :param period: Start and end years or dates of a time period,
        given as ``(YYYY, YYYY)`` or ``(YYYYMMDD, YYYYMMDD)``.
    :type period: tuple of ints
    :param str publisher: Name of a publisher.
    :param str lang: Language as a 3-letter ISO code (e.g. ``"nob"`` or ``"nno"``)
    :param str city: City of publication.
    :param str ddk: `Dewey Decimal Classification
        <https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon>`_ identifier.
    :param str topic: Topic of the documents.
    :return: a ``pandas.DataFrame`` with the resulting frequency counts of the word(s),
        spread across years. One year per row.
    """
    params = {"word": word, "title": title, "period": period, "publisher": publisher,
              "lang": lang, "city": city, "ddk": ddk, "topic": topic}
    if isinstance(word, str):
        # assume a comma separated string
        word = [w.strip() for w in word.split(",")]
    params["word"] = tuple(word)
    params = {x: params[x] for x in params if params[x] is not None}
    r = api_post(BASE_URL + "/ngram_book", json=params)
    df = pd.DataFrame.from_dict(r.json(), orient="index")
    df.index = df.index.map(lambda x: tuple(x.split()))
    if not isinstance(df.index, pd.MultiIndex):
        raise DHLabApiError(f"{isinstance(df.index, pd.MultiIndex)=}")
    columns = df.index.levels[0]
    df = pd.concat([df.loc[x] for x in columns], axis=1)
    df.columns = columns
    return df


# @_docstring_parameters_from(_ngram_doc, drop="doctype")
def ngram_periodicals(
    word: Union[List, str] = ["."],
    title: str | None = None,
    period: Tuple[int, int] | None = None,
    publisher: str | None = None,
    lang: str | None = None,
    city: str | None = None,
    ddk: str | None = None,
    topic: str | None = None,
    **kwargs,
) -> DataFrame:
    """Get a time series of frequency counts for ``word`` in periodicals.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/ngram_periodicals`.

    :param word: Word(s) to search for.
        Can be several words in a single string, separated by comma, e.g. ``"ord,ordene,orda"``.
    :type word: str or list of str
    :param str title: Title of a specific document to search through.
    :param period: Start and end years or dates of a time period,
        given as ``(YYYY, YYYY)`` or ``(YYYYMMDD, YYYYMMDD)``.
    :type period: tuple of ints
    :param str publisher: Name of a publisher.
    :param str lang: Language as a 3-letter ISO code (e.g. ``"nob"`` or ``"nno"``)
    :param str city: City of publication.
    :param str ddk: `Dewey Decimal Classification
        <https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon>`_ identifier.
    :param str topic: Topic of the documents.
    :return: a ``pandas.DataFrame`` with the resulting frequency counts of the word(s),
        spread across years. One year per row.
    """
    params = {"word": word, "title": title, "period": period, "publisher": publisher,
              "lang": lang, "city": city, "ddk": ddk, "topic": topic}
    if isinstance(word, str):
        # assume a comma separated string
        word = [w.strip() for w in word.split(",")]
    params["word"] = tuple(word)
    params = {x: params[x] for x in params if params[x] is not None}
    r = api_post(BASE_URL + "/ngram_periodicals", json=params)
    df = pd.DataFrame.from_dict(r.json(), orient="index")
    df.index = df.index.map(lambda x: tuple(x.split()))
    if not isinstance(df.index, pd.MultiIndex):
        raise DHLabApiError(f"{isinstance(df.index, pd.MultiIndex)=}")
    columns = df.index.levels[0]
    df = pd.concat([df.loc[x] for x in columns], axis=1)
    df.columns = columns
    return df


def ngram_news(
    word: Union[List, str] = ["."],
    title: str | None = None,
    period: Tuple[int, int] | None = None,
) -> DataFrame:
    """Get a time series of frequency counts for ``word`` in newspapers.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/ngram_newspapers`.

    :param word: Word(s) to search for.
        Can be several words in a single string, separated by comma, e.g. ``"ord,ordene,orda"``.
    :type word: str or list of str
    :param str title: Title of a specific newspaper to search through.
    :param period: Start and end years or dates of a time period,
        given as ``(YYYY, YYYY)`` or ``(YYYYMMDD, YYYYMMDD)``.
    :type period: tuple of ints
    :return: a ``pandas.DataFrame`` with the resulting frequency counts of the word(s),
        spread across the dates given in the time period. Either one year or one day per row.
    """
    params = {"word": word, "title": title, "period": period}
    if isinstance(word, str):
        # assume a comma separated string
        word = [w.strip() for w in word.split(",")]
    params["word"] = tuple(word)
    params = {x: params[x] for x in params if params[x] is not None}
    r = api_post(BASE_URL + "/ngram_newspapers", json=params)
    df = pd.DataFrame.from_dict(r.json(), orient="index")
    df.index = df.index.map(lambda x: tuple(x.split()))
    if not isinstance(df.index, pd.MultiIndex):
        raise DHLabApiError(f"{isinstance(df.index, pd.MultiIndex)=}")
    columns = df.index.levels[0]
    df = pd.concat([df.loc[x] for x in columns], axis=1)
    df.columns = columns
    return df

def _create_sparse_matrix(structure: dict[str, dict[str, int]]):
    """Create a sparse DataFrame from an API counts object."""

    # TODO: This should maybe be sorted
    all_words = list(set(word for dct in structure.values() for word in dct))

    # sparse_cols: {dhlabid: sparse_series, ...}
    # https://pandas.pydata.org/docs/reference/api/pandas.SparseDtype.html
    sparse_cols = {}

    for dhlabid, wordfreqs in structure.items():
        sparse_cols[dhlabid] = pd.Series(
            wordfreqs,
            index=all_words,
            dtype=int # See below
        # SparseDtype(int) does not allow NaN. Convert NaN to 0 first.
        ).fillna(0).astype(SparseDtype(int, 0))

    # Sparse DataFrame
    return pd.DataFrame(sparse_cols)

def get_document_frequencies(
    urns: List[str], cutoff: int = 0, words: List[str] | None = None, sparse: bool = False
) -> DataFrame:
    """Fetch frequency counts of ``words`` in documents (``urns``).

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/frequencies`.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param int cutoff: minimum frequency of a word to be counted
    :param list words: a list of words to be counted - if left None, whole document is returned. If not None both the counts and their relative frequency is returned.
    :param bool sparse: create a sparse matrix for memory efficiency
    """
    params = {"urns": urns, "cutoff": cutoff, "words": words}

    r = api_post(f"{BASE_URL}/frequencies", json=params)
    result = r.json()

    # check if words are passed - return differs a bit
    if words is None:
        structure = {}

        # result: [
        #   [ [dhlabid_1, word, frequency], [dhlabid_1, ...], ...]
        #   [ [dhlabid_2, word, frequency], [dhlabid_2, ...], ...]
        # ]
        for res_for_urn in result:
            try:
                # structure: {dhlabid: {word: frequency, ...}, ...}
                dhlabid = res_for_urn[0][0]
                structure[dhlabid] = {x[1]: x[2] for x in res_for_urn}
            except IndexError:
                pass

        if sparse == True:
            df = _create_sparse_matrix(structure)
        else:
            df = pd.DataFrame(structure)
        
        df = df.sort_values(by=df.columns[0], ascending=False).fillna(0)
    else:
        df = pd.DataFrame(result)
        df.columns = ["urn", "word", "freq", "urncount"]
        df["relfreq"] = df["freq"] / df.urncount
        df = pd.pivot_table(
            df, values=["freq", "relfreq"], index="word", columns="urn"
        ).fillna(0)
    return df


def get_word_frequencies(
    urns: List[str], cutoff: int = 0, words: List[str] | None = None
) -> DataFrame:
    """Fetch frequency numbers for ``words`` in documents (``urns``).

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/frequencies`.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param int cutoff: minimum frequency of a word to be counted
    :param list words: a list of words to be counted - should not be left None.
    """
    return get_document_frequencies(urns, cutoff, words)


def get_urn_frequencies(urns: List[str], dhlabid: List[int] | None = None) -> DataFrame:
    """Fetch frequency counts of documents as URNs or DH-lab ids.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/frequencies`.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param list dhlabid: list of numbers for dhlabid:
        ``[1000001, 2000003]``
    """
    if dhlabid is None:
        params = {"urns": urns}
    else:
        params = {"dhlabid": dhlabid}
    r = api_post(f"{BASE_URL}/urn_frequencies", json=params)
    result = r.json()
    # check if words are passed - return differs a bit
    df = pd.DataFrame(result)
    df.columns = ["urn", "freq"]
    return df


def document_corpus(
    doctype: str | None = None,
    author: str | None = None,
    freetext: str | None = None,
    fulltext: str | None = None,
    from_year: int | None = None,
    to_year: int | None = None,
    from_timestamp: int | None = None,
    to_timestamp: int | None = None,
    title: str | None = None,
    ddk: str | None = None,
    subject: str | None = None,
    publisher: str | None = None,
    literaryform: str | None = None,
    genres: str | None = None,
    city: str | None = None,
    lang: str | None = None,
    limit: int | None = None,
    order_by: str | None = None,
) -> DataFrame:
    """Fetch a corpus based on metadata.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/build_corpus <https://api.nb.no/dhlab/#/default/post_build_corpus>`_.

    At least one filter argument must be set.

    :param str doctype: ``"digibok"``, ``"digavis"``, ``"digitidsskrift"`` or ``"digistorting"``
    :param str author: Name of an author.
    :param str freetext: any of the parameters, for example: ``"digibok AND Ibsen"``.
    :param str fulltext: words within the publication.
    :param int from_year: Start year for time period of interest.
    :param int to_year: End year for time period of interest.
    :param int from_timestamp: Start date for time period of interest.
        Format: ``YYYYMMDD``, books have ``YYYY0101``
    :param int to_timestamp: End date for time period of interest.
        Format: ``YYYYMMDD``, books have ``YYYY0101``
    :param str title: Name or title of a document.
    :param str ddk: `Dewey Decimal Classification
        <https://no.wikipedia.org/wiki/Deweys_desimalklassifikasjon>`_ identifier.
    :param str subject: subject (keywords) of the publication.
    :param str publisher: Name of publisher.
    :param str literaryform: literary form of the publication (books)
    :param str genres: genre of the publication.
    :param str city: place of publication
    :param str lang: Language of the publication, as a 3-letter ISO code.
        Example: ``"nob"`` or ``"nno"``
    :param int limit: number of items to sample.
    :param str order_by: order of elements in the corpus object. Typically used in combination with a limit. Example ``"random"`` (random order, the slowest), ``"rank"`` (ordered by relevance, faster) or ``"first"`` (breadth-first, using the order in the database table, the fastest method)
    :return: a ``pandas.DataFrame`` with the corpus information.
    """
    parms = {"doctype": doctype, "author": author, "freetext": freetext,
             "fulltext": fulltext, "from_year": from_year, "to_year": to_year,
             "from_timestamp": from_timestamp, "to_timestamp": to_timestamp,
             "title": title, "ddk": ddk, "subject": subject, "publisher": publisher,
             "literaryform": literaryform, "genres": genres, "city": city,
             "lang": lang, "limit": limit, "order_by": order_by}

    params = {x: parms[x] for x in parms if parms[x] is not None}

    if not any(params.values()):
        raise ValueError("`document_corpus` needs at least one filter argument.")

    r = api_post(BASE_URL + "/build_corpus", json=params)

    return pd.DataFrame(r.json())


get_document_corpus = document_corpus # Function alias


def urn_collocation(
    urns: List[str] | None = None,
    word: str = "arbeid",
    before: int = 5,
    after: int = 0,
    samplesize: int = 200000,
) -> DataFrame:
    """Create a collocation from a list of URNs.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/urncolldist_urn`.

    :param list urns: list of uniform resource name strings, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param str word: word to construct collocation with.
    :param int before: number of words preceding the given ``word``.
    :param int after: number of words following the given ``word``.
    :param int samplesize: total number of ``urns`` to search through.
    :return: a ``pandas.DataFrame`` with distance (sum of distances and bayesian distance) and
        frequency for words collocated with ``word``.
    """

    params = {
        "urn": urns,
        "word": word,
        "before": before,
        "after": after,
        "samplesize": samplesize,
    }
    r = api_post(BASE_URL + "/urncolldist_urn", json=params)
    return pd.read_json(StringIO(r.json()))


def totals(top_words: int = 50000) -> DataFrame:
    """Get aggregated raw frequencies of all words in the National Library's database.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/totals/{top_words} <https://api.nb.no/dhlab/#/default/get_totals__top_words_>`_.

    :param int top_words: The number of words to get total frequencies for.
    :return: a ``pandas.DataFrame`` with the most frequent words.
    """
    resp = api_get(BASE_URL + f"/totals/{top_words}")

    return pd.DataFrame.from_dict(dict(resp.json()), orient="index", columns=["freq"])


def concordance(
    urns: list | None = None, words: str | None = None, window: int = 25, limit: int = 100
) -> DataFrame:
    """Get a list of concordances from the National Library's database.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/conc <https://api.nb.no/dhlab/#/default/post_conc>`_.

    :param list urns: uniform resource names, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param str words: Word(s) to search for.
        Can be an SQLite fulltext query, an fts5 string search expression.
    :param int window: number of tokens on either side to show in the collocations, between 1-25.
    :param int limit: max. number of concordances per document. Maximum value is 1000.
    :return: a table of concordances
    """
    if words is None:
        return pd.DataFrame(columns=["index", "docid", "urn", "conc"])  # exit condition
    else:
        params = {"urns": urns, "query": words, "window": window, "limit": limit}
        r = api_post(BASE_URL + "/conc", json=params)
    return pd.DataFrame(r.json())


konkordans = concordance # Function alias


def concordance_counts(
    urns: list | None = None, words: str | None = None, window: int = 25, limit: int = 100
) -> DataFrame:
    """Count concordances (keyword in context) for a corpus query (used for collocation analysis).

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/conccount <https://api.nb.no/dhlab/#/default/post_conccount>`_.

    :param list urns: uniform resource names, for example:
        ``["URN:NBN:no-nb_digibok_2008051404065", "URN:NBN:no-nb_digibok_2010092120011"]``
    :param str words: Word(s) to search for.
        Can be an SQLite fulltext query, an fts5 string search expression.
    :param int window: number of tokens on either side to show in the collocations, between 1-25.
    :param int limit: max. number of concordances per document. Maximum value is 1000.
    :return: a table of counts
    """
    if words is None:
        return pd.DataFrame(columns=["freq"])  # exit condition
    else:
        params = {"urns": urns, "query": words, "window": window, "limit": limit}
        r = api_post(BASE_URL + "/conccount", json=params)
    return pd.DataFrame(r.json())


def word_concordance(
    urn: list | None = None,
    dhlabid: list | None = None,
    words: list | None = None,
    before: int = 12,
    after: int = 12,
    limit: int = 100,
    samplesize: int = 50000,
) -> DataFrame:
    """Get a list of concordances from the National Library's database.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint
    `/conc <https://api.nb.no/dhlab/#/default/conc_word_urn>`_.

    :param list urns: dhlab serial ids. (server can take both urns and dhlabid but so we may rewrite this to)
    :param str words: Word(s) to search for -- must be a list
    :param int before: between 0-24.
    :param int after: between 0-24 (before + sum <= 24)
    :param int limit: max. number of concordances per server process.
    :param int samplesize: samples from urns.
    :return: a table of concordances
    """

    # server checks if either dhlabid or urns are present in the parameters, so only one of them
    # is passed. The return is dhlabid anyhow.

    if dhlabid is not None:
        params = {
            "dhlabid": dhlabid,
            "words": words,
            "before": before,
            "after": after,
            "limit": limit,
            "samplesize": samplesize,
        }
    elif urn is not None:
        params = {
            "urn": urn,
            "words": words,
            "before": before,
            "after": after,
            "limit": limit,
            "samplesize": samplesize,
        }
    else:
        params = {
            "words": words,
            "before": before,
            "after": after,
            "limit": limit,
            "samplesize": samplesize,
        }

    r = api_post(BASE_URL + "/conc_word_urn", json=params)

    return pd.DataFrame(
        [x for y in r.json() for x in y],
        columns=["dhlabid", "before", "target", "after"],
    )


def collocation(
    corpusquery: str = "norge", word: str = "arbeid", before: int = 5, after: int = 0
) -> DataFrame:
    """Make a collocation from a corpus query.

    :param str corpusquery: query string
    :param str word: target word for the collocations.
    :param int before: number of words prior to ``word``
    :param int after: number of words following ``word``
    :return: a dataframe with the resulting collocations
    """
    params = {
        "metadata_query": corpusquery,
        "word": word,
        "before": before,
        "after": after,
    }
    r = api_post(BASE_URL + "/urncolldist", json=params)
    return pd.read_json(r.json())


# Norwegian word bank


def word_variant(word: str, form: str, lang: str = "nob") -> list:
    """Find alternative ``form`` for a given ``word`` form.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint ``/variant_form``

    Example: ``word_variant('spiste', 'pres-part')``

    :param str word: any word string
    :param str form: a morphological feature tag from the Norwegian wordbank
        `"Orbanken" <https://www.nb.no/sprakbanken/ressurskatalog/oai-nb-no-sbr-5/>`_.
    :param str lang: either "nob" or "nno"
    """
    resp = api_get(
        f"{BASE_URL}/variant_form",
        params={"word": word, "form": form, "lang": lang}
    )

    return resp.json()


def word_paradigm(word: str, lang: str = "nob") -> list:
    """Find paradigms for a given ``word`` form.

    Call the API :py:obj:`~dhlab.constants.BASE_URL` endpoint ``/paradigm``

    Example:

    .. code-block:: python

        word_paradigm('spiste')
        # [['adj', ['spisende', 'spist', 'spiste']],
        # ['verb', ['spis', 'spise', 'spiser', 'spises', 'spist', 'spiste']]]

    :param str word: any word string
    :param str lang: either "nob" or "nno"
    """
    resp = api_get(
        f"{BASE_URL}/paradigm",
        params={"word": word, "lang": lang}
    )

    return resp.json()


def word_paradigm_many(wordlist: list, lang: str = "nob") -> list:
    """Find alternative forms for a list of words.
    :param wordlist: `List` of words
    :param lang: Language
    """
    r = api_post(f"{BASE_URL}/paradigms", json={"words": wordlist, "lang": lang})
    return r.json()


def word_form(word: str, lang: str = "nob") -> list:
    """Look up the morphological feature specification of a ``word`` form.
    :param word: Word
    :param lang: Language
    """
    resp = api_get(
        f"{BASE_URL}/word_form",
        params={"word": word, "lang": lang}
    )

    return resp.json()


def word_form_many(wordlist: list, lang: str = "nob") -> list:
    """Look up the morphological feature specifications for word forms in a ``wordlist``.
    :param wordlist: `List` of words
    :param lang: Language
    """
    r = api_post(f"{BASE_URL}/word_forms", json={"words": wordlist, "lang": lang})
    return r.json()


def word_lemma(word: str, lang: str = "nob") -> list:
    """Find the list of possible lemmas for a given ``word`` form.
    :param word: Word to find lemmas for
    :param lang: Language
    """
    r = api_get(
        f"{BASE_URL}/word_lemma",
        params={"word": word, "lang": lang}
    )

    return r.json()


def word_lemma_many(wordlist, lang="nob"):
    """Find lemmas for a list of given word forms."""


def query_imagination_corpus(
    category=None,
    author=None,
    title=None,
    year=None,
    publisher=None,
    place=None,
    oversatt=None,
):
    """Fetch data from imagination corpus"""
    params = {"category": category, "author": author, "title": title, "year": year,
              "publisher": publisher, "place": place, "oversatt": oversatt}
    params = {key: params[key] for key in params if params[key] is not None}

    resp = api_get(f"{BASE_URL}/imagination", params=params)
    return resp.json()



from typing import Union

import pandas as pd

from dhlab.api.dhlab_api import (
    word_form,
    word_form_many,
    word_lemma,
    word_lemma_many,
    word_paradigm,
    word_paradigm_many,
)


class WordbankSuper:
    """Super class for wordbank classes"""

    def __init__(self, frame):
        self.frame = frame

    def __repr__(self) -> str:
        return self.frame.__repr__()

    def _repr_html_(self) -> Union[str, None]:
        """
        Return the HTML representation of the DhlabObj frame attribute
        """
        return self.frame._repr_html_()


class WordParadigm(WordbankSuper):
    """Fetch inflection paradigms for a list of words, or just one word"""

    def __init__(self, words, lang="nob"):
        if isinstance(words, list):
            self.paradigms = pd.DataFrame(word_paradigm_many(words, lang))
        elif isinstance(words, str):
            self.paradigms = pd.DataFrame(word_paradigm(words, lang))
        else:
            self.paradigms = None

        super().__init__(self.paradigms)


class WordForm(WordbankSuper):
    """Fetch possible forms of a word or list of words"""

    def __init__(self, words, lang="nob"):
        if isinstance(words, list):
            self.forms = pd.DataFrame(word_form_many(words, lang))
        elif isinstance(words, str):
            self.forms = pd.DataFrame(word_form(words, lang))
        else:
            self.forms = None

        super().__init__(self.forms)


class WordLemma(WordbankSuper):
    """Fetch possbile lemmas for a given word form"""

    def __init__(self, words, lang="nob"):
        if isinstance(words, list):
            self.lemmas = pd.DataFrame(word_lemma_many(words, lang))
        elif isinstance(words, str):
            self.lemmas = pd.DataFrame(word_lemma(words, lang))
        else:
            self.lemmas = None

        super().__init__(self.lemmas)



import json

import networkx as nx
import requests

from dhlab.constants import GALAXY_API, NGRAM_API
from dhlab.api.utils import api_get


def get_ngram(terms: str, corpus: str = "avis", lang: str = "nob", session: requests.Session | None = None) -> dict:
    """Fetch raw and relative frequencies for the ``terms``.

    Call the :py:data:`NGRAM_API`.
    The frequencies are aggregated per year between 1800-2021.

    :param str terms: comma separated string of words
    :param str corpus: type of documents to search through
    :return: table of annual frequency counts per term
    """

    resp = api_get(
        NGRAM_API,
        params={"terms": terms, "corpus": corpus, "lang": lang},
        session=session
    )

    return json.loads(resp.text)


def make_word_graph(
    words: str, corpus: str = "all", cutoff: int = 16, leaves: int = 0, session: requests.Session | None = None
) -> nx.DiGraph:
    """Get galaxy from ngram-database.

    Call the :py:obj:`~dhlab.constants.GALAXY_API` endpoint.

    :param str words: comma-separated string of words
    :param str corpus: document type: ``'book'``, ``'avis'``, or ``'all'``,
    :param int cutoff: Number of nodes to include.
    :param int leaves: Set leaves=1 to get the leaves.
    :return: A `networkx.DiGraph` with the results.
    """
    params = dict()
    params["terms"] = words
    params["corpus"] = corpus
    params["limit"] = cutoff
    params["leaves"] = leaves

    resp = api_get(GALAXY_API, params=params, session=session)

    G = nx.DiGraph()
    edgelist = []

    graph = json.loads(resp.text)
    nodes = graph["nodes"]
    edges = graph["links"]

    for edge in edges:
        edgelist += [
            (
                nodes[edge["source"]]["name"],
                nodes[edge["target"]]["name"],
                abs(edge["value"]),
            )
        ]

    G.add_weighted_edges_from(edgelist)

    return G



import re
import pandas as pd
from pandas import DataFrame
from typing import List
from dhlab.api.dhlab_api import (
    concordance,
    get_document_frequencies,
    urn_collocation,
    word_concordance,
    concordance_counts,
)
from dhlab.text.dhlab_object import DhlabObj
from dhlab.text.corpus import Corpus
from dhlab.text.utils import urnlist


# convert cell to a link
def make_link(row):
    r = "<a target='_blank' href = 'https://urn.nb.no/{x}'>{x}</a>".format(x=str(row))
    return r


# find hits a cell
def find_hits(x):
    return " ".join(re.findall("<b>(.+?)</b", x))


class Concordance(DhlabObj):
    """Wrapper for concordance function"""

    def __init__(
        self,
        corpus: Corpus | None = None,
        query: str | None = None,
        window: int = 20,
        limit: int = 500
):
        """Get concordances for word(s) in corpus

        :param corpus: Target corpus, defaults to None
        :param query: word or list or words, defaults to None
        :param window: how many tokens to consider around the target word, \
            defaults to 20
        :param limit: limit returned hits, defaults to 500
        """

        if corpus is None:
            self.concordance = pd.DataFrame()
            self.corpus = None
        else:
            self.concordance = concordance(
                urns=urnlist(corpus), words=query, window=window, limit=limit
            )
            self.concordance["link"] = self.concordance.urn.apply(make_link)
            self.concordance = self.concordance[["link", "urn", "conc"]]
            self.concordance.columns = ["link", "urn", "concordance"]
            self.corpus = corpus

        super().__init__(self.concordance)

    def show(self, n: int = 10, style: bool = True):
        result = self.concordance.sample(min(n, self.size))

        if style:
            return result[["link", "concordance"]].style
        else:
            return result

    @classmethod
    def from_df(cls, df: pd.DataFrame):
        "Typecast DataFrame to Concordance"
        obj = Concordance()
        obj.concordance = df
        obj.frame = df
        return obj


class Collocations(DhlabObj):
    """Collocations"""

    def __init__(
        self,
        corpus: Corpus | None = None,
        words: str | list[str] | None = None,
        before: int = 10,
        after: int = 10,
        reference: pd.DataFrame | None = None,
        samplesize: int = 20000,
        alpha: bool = False,
        ignore_caps: bool = False,
    ):
        """Create collocations object

        :param corpus: target corpus, defaults to None
        :type corpus: dh.Corpus, optional
        :param words: target words(s), defaults to None
        :type words: str or list, optional
        :param before: words to include before, defaults to 10
        :type before: int, optional
        :param after: words to include after, defaults to 10
        :type after: int, optional
        :param reference: reference frequency list, defaults to None
        :type reference: pd.DataFrame, optional
        :param samplesize: _description_, defaults to 20000
        :type samplesize: int, optional
        :param alpha: Only include alphabetical tokens, defaults to False
        :type alpha: bool, optional
        :param ignore_caps: Ignore capitalized letters, defaults to False
        :type ignore_caps: bool, optional
        """
        if isinstance(words, str):
            words = [words]

        if corpus is not None and words is not None:
            coll = pd.concat(
                [
                    urn_collocation(
                        urns=urnlist(corpus),
                        word=w,
                        before=before,
                        after=after,
                        samplesize=samplesize,
                    )
                    for w in words
                ]
            )[["counts"]]
        else:
            coll = pd.DataFrame()

        if alpha:
            coll = coll.loc[[x for x in coll.index if x.isalpha()]]
            if reference is not None:
                reference = reference.loc[[x for x in reference.index if x.isalpha()]]

        if ignore_caps:
            coll.index = [x.lower() for x in coll.index]
            if reference is not None:
                reference.index = [x.lower() for x in reference.index]

        self.coll = coll.groupby(coll.index).sum()
        self.reference = reference
        self.before = before
        self.after = after

        if self.reference is not None:
            teller = self.coll.counts / self.coll.counts.sum()
            divisor = self.reference.iloc[:, 0] / self.reference.iloc[:, 0].sum()
            self.coll["relevance"] = teller / divisor

        super().__init__(self.coll)

    def show(self, sortby: str = "counts", n: int = 20):
        return self.coll.sort_values(by=sortby, ascending=False).head(n)

    def keywordlist(self, top: int = 200, counts: int = 5, relevance: float = 10):
        mask = self.coll[self.coll.counts > counts]
        mask = mask[mask.relevance > relevance]
        return list(mask.sort_values(by="counts", ascending=False).head(200).index)

    @classmethod
    def from_df(cls, df: pd.DataFrame):
        """Typecast DataFrame to Collocation

        :param df: DataFrame
        :return: Collocation
        """
        obj = Collocations()
        obj.coll = df
        obj.frame = df
        return obj


class Counts(DhlabObj):
    """Provide counts for a corpus - shouldn't be too large"""

    def __init__(
        self,
        corpus: Corpus | None = None,
        words: list[str] | None = None,
        cutoff: int = 0,
        sparse: bool = True
):
        """Get frequency list for Corpus

        :param corpus: target Corpus, defaults to None
        :param words: list of words to be counted, defaults to None
        :param cutoff: frequency cutoff, will not include words with frequency < cutoff
        :param sparse: return a sparse matrix for memory efficiency
        """
        if corpus is None and words is None:
            self.freq = pd.DataFrame()
            self.title_dct = None
        elif corpus is not None:
            # Make sure corpus is a dhlab corpus
            # if not, try to make it one
            # if isinstance(corpus, pd.DataFrame):
            #     corpus = dh.Corpus.from_df(corpus)

            # if not isinstance(corpus, dh.Corpus):
            #     raise TypeError("Corpus must be of type dh.Corpus or pd.DataFrame")

            # count - if words is none result will be as if counting all words
            # in the corpus
            self.freq = get_document_frequencies(
                urns=urnlist(corpus), cutoff=cutoff, words=words, sparse=sparse
            )

            # Include dhlab and title link in object
            try:
                self.title_dct = {
                    k: v for k, v in zip(corpus.frame.dhlabid, corpus.frame.title)
                }
            except:
                self.title_dct = None

            # Add relative frequencies if available
            if words is not None:
                self.relfreq = self.freq.relfreq
                self.freq = self.freq.freq

        super().__init__(self.freq)

    def is_sparse(self):
        """Function to report sparsity of counts frame"""
        try:
            density = self.freq.sparse.density
            if density:
                sparse = True
            else:
                sparse = False
        except:
            sparse = False
        return sparse

    def sum(self):
        """Summarize Corpus frequencies
        :return: frequency list for Corpus
        """

        # Needed since Pandas seems to make sparse matrices dense when summing
        if self.is_sparse() == True:
            # convert to coo matrix for iteration
            coo_matrix = self.freq.sparse.to_coo()

            # get the words and their indices
            rowidx = {idx: word for idx, word in enumerate(list(self.freq.index))}

            # build a freq dictionary by looping
            freqDict = dict()

            for i,j,v in zip(coo_matrix.row, coo_matrix.col, coo_matrix.data):
                word = rowidx[i]
                if word in freqDict:
                    freqDict[word] += v
                else:
                    freqDict[word] = v

            df = pd.DataFrame(freqDict.items(), columns=["word", "freq"]).set_index("word").sort_values(by="freq", ascending=False)
            df.index.name = None
            return self.from_df(df)
        else:
            return self.from_df(self.counts.sum(axis=1).to_frame("freq"))

    def display_names(self):
        "Display data with record names as column titles."
        assert self.title_dct is not None, "No titles available"
        return self.frame.rename(self.title_dct, axis=1)

    def display_rel_names(self):
        "Display relfreq data with record names as column titles."
        return self.relfreq.rename(self.title_dct, axis=1)

    @classmethod
    def from_df(cls, df: pd.DataFrame):
        obj = Counts()
        obj.freq = df
        obj.frame = df
        return obj

    ### Legacy properties and methods ###

    @property
    def counts(self):
        "Legacy property for freq"
        return self.freq


class WordConc(DhlabObj):
    def __init__(
        self,
        frame: DataFrame | None = None,
        urn: list | None = None,
        dhlabid: list | None = None,
        words: List[str] | None = None,
        before: int = 12,
        after: int = 12,
        limit: int = 100,
        samplesize: int = 50000,
    ):
        """Wrapper for word_concordance function

        Args:
            frame (DataFrame, optional): Use to typecast Dataframe with data to this class. Defaults to None.
            urn (list, optional): NB URN identifiers for target text objects. Defaults to None.
            dhlabid (list, optional): DHLAB identifiers for target books. Defaults to None.
            words (List[str], optional): Target words. Defaults to None.
            before (int, optional): Window of tokens before target word to return. Defaults to 12.
            after (int, optional): Window of tokens after target word to return. Defaults to 12.
            limit (int, optional): Limit returned results. Defaults to 100.
            samplesize (int, optional): Size of sample. Defaults to 50000.
        """
        params = {
            "urn": urn,
            "dhlabid": dhlabid,
            "words": words,
            "before": before,
            "after": after,
            "limit": limit,
            "samplesize": samplesize,
        }

        if (
            frame is None
            and words is not None
            and (urn is not None or dhlabid is not None)
        ):
            frame = word_concordance(**params)

        if frame is None:
            frame = DataFrame()

        super().__init__(frame)


class ConcCounts(DhlabObj):
    def __init__(
        self,
        frame: DataFrame | None = None,
        urn: list | None = None,
        dhlabid: list | None = None,
        words: str | None = None,
        window: int = 25,
        limit: int = 100,
    ):
        """Wrapper for concordance_counts function

        Args:
            frame (DataFrame, optional): Concordance data. Use to typecast to CountCounts. Defaults to None.
            urn (list, optional): NB URN identifiers. Defines which objects to search. Defaults to None.
            dhlabid (list, optional): DHLAB identifiers. Defines which objects to search. Defaults to None.
            words (str, optional): String with SQLite ft search with target words. Defaults to None.
            window (int, optional): text window around target words to return. Defaults to 25.
            limit (int, optional): Number of results to return. Defaults to 100.
        """
        params = {
            "urns": urn,
            "words": words,
            "window": window,
            "limit": limit,
        }

        if dhlabid is not None:
            print("Dhlabid not (yet) supported for concordance_counts")

        if isinstance(words, list):
            params["words"] = " OR ".join(words)
            print(
                f"words supports a SQLite fulltext query. Converting to 'OR' query: {params['words']}"
            )

        if frame is None and words is not None and urn is not None:
            frame = concordance_counts(**params)

        if frame is None:
            frame = DataFrame()

        super().__init__(frame)



from urllib.parse import urlparse

import requests
from IPython.display import HTML, display


def code_toggle(button_text="Klikk for å vise/skjule kodeceller"):
    display(
        HTML(
            """<div>
                <style>
                 .mybutton {
                    background-color: lightgrey;
                    border: none;
                    color: white;
                    padding: 10px 16px;
                    text-align: center;
                    text-decoration: none;
                    display: inline-block;
                    font-size: 16px;
                    margin: 4px 2px;
                    cursor: pointer;
                }
            </style>
            <script>
                code_show=true;
                function code_toggle() {
                 if (code_show){
                 $('div.input').hide();
                 } else {
                 $('div.input').show();
                 }
                 code_show = !code_show
                }
                $( document ).ready(code_toggle);
            </script>
            <form  action="javascript:code_toggle()">
                <input class='mybutton' type="submit" value="""
            + '"'
            + button_text
            + '"'
            + """>
            </form>
        </div>"""
        )
    )


def css(
    url="https://raw.githubusercontent.com/Yoonsen/Modules/master/css_style_sheets/nb_notebook.css",
):
    """Associate a css stylesheet with the notebook,
    just specify a file or web reference, default is a custom css"""

    uri = urlparse(url)
    css_file = ""

    if uri.scheme.startswith("http"):
        query = requests.get(url)
        if query.status_code == 200:
            css_file = query.text

    elif uri.scheme == "file":
        # assume on form "file:/// on windows there is drive letter on unix not"
        file_path = url[7:]
        if file_path[2] == ":":  # then windows drive reference
            file_path = file_path[1:]
        with open(file_path, encoding="utf-8") as file:
            css_file = file.read()
    else:
        # assume string is a file locator
        with open(url, encoding="utf-8") as file:
            css_file = file.read()

    return HTML(f"<style>{css_file}</style>")



import requests

class DHLabApiError(requests.exceptions.HTTPError):
    pass

def validate_response_status(response: requests.Response) -> None:
    if response.status_code != 200:
        raise DHLabApiError(
            f"Status code: {response.status_code}: {response.reason}\n\n"
            + "There is an error connecting to the DHLab API. This is most likely not an error in your code,"
            + " but with the DHLab library or API. Try running the code again, and if that doesn't work, try to"
            + " reinstall/update the dhlab package (e.g. by running: 'pip install --upgrade dhlab').\nIf the error"
            + " persists, then please leave an issue at https://github.com/NationalLibraryOfNorway/DHLAB/issues/."
        )

def api_request(
    method: str,
    url: str,
    params: dict | None = None,
    json: dict | None = None,
    session: requests.Session | None = None
):
    if session is None:
        session = requests.Session()

    res = session.request(method, url, params=params, json=json)
    validate_response_status(res)

    return res

def api_get(url: str, params: dict | None = None, session: requests.Session | None = None):
    return api_request("GET", url, params=params, session=session)

def api_post(url: str, json: dict | None = None, session: requests.Session | None = None):
    return api_request("POST", url, json=json, session=session)




import os

"""Constant variables used across modules, override using environment variable"""

BASE_URL = os.getenv("NB_DHLAB_BASE_URL", "https://api.nb.no/dhlab")  #: REST-API URL adress to fulltext query functions
NGRAM_API =  os.getenv("NB_DHLAB_NGRAM_API", "https://api.nb.no/dhlab/nb_ngram/ngram/query") #: URL adress for API calls to ngram-databases
GALAXY_API = os.getenv("NB_DHLAB_GALAXY_API", "https://api.nb.no/dhlab/nb_ngram_galaxies/galaxies/query")  #: URL adress for word galaxy API queries



from typing import Iterable

import pandas as pd
import requests


def load_picture(url: str):
    """Load the raw image object from a URL."""
    r = requests.get(url, stream=True)
    r.raw.decode_content = True
    return r.raw


def iiif_manifest(urn: str):
    """Fetch the IIIF manifest of the scanned book images"""
    r = requests.get(f"https://api.nb.no/catalog/v1/iiif/{urn}/manifest")
    return r.json()


def mods(urn: str):
    r = requests.get(f"https://api.nb.no:443/catalog/v1/metadata/{urn}/mods")
    return r.json()


def super_search(
    term: str, number: int = 50, page: int = 0, mediatype: str = "bilder"
) -> dict:
    """Search for a term, return a json object with the matches.

    :param str term: Search term, word, or token. Default is the empty string.
    :param int number: Size of the response. Maximum 50.
    :param int page: The page number to search from.
        .. todo:: Fill in description.
    :param str mediatype: Default 'bilder'.
        .. todo::  Which other value options are available for this parameter?
    :return: a dict (json object) with the search results.
    """
    url = "https://api.nb.no:443/catalog/v1/items"
    number = min(number, 50)
    if term == "":
        r = requests.get(
            url,
            params={"filter": f"mediatype:{mediatype}", "page": page, "size": number},
        )
    else:
        r = requests.get(
            url,
            params={
                "q": term,
                "filter": f"mediatype:{mediatype}",
                "page": page,
                "size": number,
            },
        )
    return r.json()


def total_search(size: int = 50, page: int = 0) -> dict:
    """Retrieve the first ``size`` occurrences from ``page``.

    Wrapper-function for :func:`super_search` with arguments
    ``medietype="bilder"`` and ``term=''``.
    """
    return super_search(term="", number=size, page=page, mediatype="bilder")


def get_df(phrases: Iterable, title: str = "aftenposten"):
    """Get ``phrases`` from documents of ``title``, and aggregate their frequencies."""
    querystring = " + ".join(['"' + frase + '"' for frase in phrases])
    query = {
        "q": querystring,
        "size": 1,
        "aggs": "year",
        "filter": f"title:{title}",
    }
    r = requests.get("https://api.nb.no/catalog/v1/items", params=query)
    aggs = r.json()["_embedded"]["aggregations"][0]["buckets"]
    return {x["key"]: x["count"] for x in aggs}


def get_json(phrases, mediatype="aviser"):
    querystring = " + ".join(['"' + frase + '"' for frase in phrases])
    query = {
        "q": querystring,
        "size": 1,
        "snippets": mediatype,
        "aggs": "year",
        "searchType": "FULL_TEXT_SEARCH",
    }
    r = requests.get("https://api.nb.no/catalog/v1/items", params=query)
    aggs = r.json()
    return aggs


def get_data(frase, media="avis", title="jazznytt"):
    query = {
        "q": '"' + frase + '""',
        "size": 1,
        "aggs": "year",
        "filter": "mediatype:{mt}".format(mt=media),
        "filter": "title:{title}".format(title=title),
    }
    r = requests.get("https://api.nb.no/catalog/v1/items", params=query)
    return r.json()


def get_data_and(frases, title="aftenposten", media="avis"):
    querystring = " + ".join(['"' + frase + '"' for frase in frases])
    print(querystring)
    query = {
        "q": querystring,
        "size": 1,
        "aggs": "year",
        "filter": "title:{title}".format(title=title),
    }
    r = requests.get("https://api.nb.no/catalog/v1/items", params=query)
    return r.json()


def get_df_pd(frase, media="bøker"):
    return pd.DataFrame.from_dict(
        get_df(frase, media=media), orient="index"
    ).sort_index()


def get_konks(urn, phrase, window=1000, n=1000):
    querystring = '"' + phrase + '"'
    query = {"q": querystring, "fragments": n, "fragSize": window}
    r = requests.get(
        "https://api.nb.no/catalog/v1/items/{urn}/contentfragments".format(urn=urn),
        params=query,
    )
    res = r.json()
    results = []
    try:
        for x in res["contentFragments"]:
            urn = x["pageid"]
            hit = x["text"]
            splits = hit.split("<em>")
            s2 = splits[1].split("</em>")
            before = splits[0]
            word = s2[0]
            after = s2[1]
            results.append({"urn": urn, "before": before, "word": word, "after": after})
    except BaseException:
        results = []
    return results


def get_phrase_info(urn, phrase, window=1000, n=1000):
    querystring = '"' + phrase + '"'
    query = {
        "q": querystring,
    }
    r = requests.get(
        f"https://api.nb.no/catalog/v1/items/{urn}/contentfragments", params=query
    )
    res = r.json()
    return res



from dhlab.api.dhlab_api import wildcard_search
from dhlab.text.dhlab_object import DhlabObj


class WildcardWordSearch(DhlabObj):
    """
    Find a class of words matching a wildcard string
    """

    def __init__(self, word, factor=2, freq_limit=10, limit=50):
        """
        Args:
            word: word from a mixture of * and characters
            factor: the additional length of words to be returned
            freq_limit: the frequency of returned words lower limit
            limit: number of words returned
        """
        self.words = wildcard_search(
            word, factor=factor, freq_limit=freq_limit, limit=limit
        )
        super().__init__(self.words)
